import logging
from abc import ABC, abstractmethod
from typing import Dict, Union, Type

from logux.core import ActionCommand, ChannelCommand, UnknownAction, UnknownSubscription

logger = logging.getLogger(__name__)


class BaseActionDispatcher(ABC):
    """ Meta class for all Dispatchers """

    @abstractmethod
    def register(self, action: ActionCommand):
        """ Add Action handler for Dispatcher """
        raise NotImplementedError()


class DefaultActionDispatcher(BaseActionDispatcher):
    """ Default logux Dispatcher for Actions """
    _actions: Dict[str, Type[ActionCommand]] = {}

    def __str__(self):
        return ', '.join(self._actions)

    def __getitem__(self, action_type: str):
        return self._actions[action_type]

    def _action_is_valid(self, action: Type[ActionCommand]) -> bool:
        if not action.action_type:
            raise ValueError('`action_type` attribute is required for all Actions')

        if self.has_action(action.action_type):
            raise ValueError(f'`{action.action_type}` action type already registered')

        if getattr(action.access, '__isabstractmethod__', False):
            raise ValueError('`access` method is required')

        return True

    def has_action(self, action_type: str) -> bool:
        """ Check if Dispatcher has handler for particular action type """
        return action_type in self._actions

    def register(self, action: Type[ActionCommand]):  # type: ignore # noqa
        if self._action_is_valid(action):
            logger.info('registering action `%s`', action.action_type)
            self._actions[action.action_type] = action


class DefaultChannelDispatcher(BaseActionDispatcher):
    """ Default logux Dispatcher for Channels """
    _subs: Dict[str, Type[ChannelCommand]] = {}

    def __str__(self):
        return ', '.join(self._subs)

    def __getitem__(self, item: str) -> Union[Type[ChannelCommand], Type[UnknownAction]]:
        for sub in self._subs.values():
            if sub.is_match(channel=item):
                return sub

        logger.warning("can't match channel name: %s", item)

        return UnknownSubscription

    def has_subscription(self, channel_pattern: str) -> bool:
        """ Check if Dispatcher has handler for a particular channel subscription type """
        return channel_pattern in self._subs

    def _sub_is_valid(self, sub) -> bool:
        if not sub.channel_pattern:
            raise ValueError('`action_type` attribute is required for all Actions')

        if self.has_subscription(sub.channel_pattern):
            raise ValueError(f'subscription for channel `{sub.channel_pattern}` already registered')

        if getattr(sub.access, '__isabstractmethod__', False):
            raise ValueError('`access` method is required')

        if getattr(sub.load, '__isabstractmethod__', False):
            raise ValueError('`load` method is required')

        return True

    def register(self, action: Type[ChannelCommand]):  # type: ignore # noqa
        if self._sub_is_valid(action) and action.channel_pattern is not None:
            logger.info('registering subscription for `%s`', action.channel_pattern)
            self._subs[action.channel_pattern] = action


class DefaultDispatcher:
    """ Shortcut for actions and channels Dispatchers """

    def __init__(self):
        self.actions = DefaultActionDispatcher()
        self.channels = DefaultChannelDispatcher()


logux = DefaultDispatcher()

__all__ = ['logux']



from django.apps import AppConfig

from logux import settings
from logux.utils import autodiscover


class LoguxConfig(AppConfig):
    """ Logux app conf """
    name = 'logux'
    verbose_name = 'Logux'

    def ready(self):
        # check if all required settings is defined
        settings.get_config()
        # import all logux_actions.py and logux_subscriptions.py from consumer modules
        autodiscover()



import json
import logging
from itertools import chain
from typing import List, Tuple, Union

from django.http import JsonResponse, HttpRequest, HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods

from logux import settings
from logux.core import AuthCommand, LoguxValue, UnknownAction, Command, LOGUX_SUBSCRIBE, \
    protocol_version_is_supported
from logux.dispatchers import logux
from logux.exceptions import LoguxProxyException

logger = logging.getLogger(__name__)


class LoguxRequest:
    """ LoguxValue is class for deserialized request from Logux Server proxy

    The constructor should extract common fields like `version` and `secret` and parse list of commands.

    By default, command parser will provide only AuthCommand implementation (with logux_auth function
    injection). Other Action should by parsed by consumer dispatcher.

    TODO: send 403 if proxy secret is wrong
    TODO: send 429 if brute force check is fail
    """

    class CommandType:
        """ All possible Logux command types.
        https://logux.io/protocols/backend/spec/#requests """
        AUTH = 'auth'
        ACTION = 'action'

        choices = [AUTH, ACTION]

    def __init__(self, request: HttpRequest):
        """ Construct the Command and check protocol version support.

        :param request: request with command from Logux Proxy
        :raises: base Exception if request protocol version is not supported by backend
        """
        try:
            self._get_body(request)
        except (TypeError, ValueError) as err:
            logger.warning('Wrong body: %s', err)
            raise LoguxProxyException('Wrong body')

        if not protocol_version_is_supported(self.version):
            logger.warning('Unsupported protocol version: %s', self.version)
            raise LoguxProxyException('Back-end protocol version is not supported')

        self.commands: List[Command] = self._parse_commands()

    def _get_body(self, request: HttpRequest):
        _body = json.loads(request.body.decode('utf-8'))

        self.version: int = int(_body['version'])
        self.secret: str = _body['secret']
        self.raw_commands = _body['commands']

    def _parse_commands(self) -> List[Command]:
        commands: List[Command] = []

        for cmd in self.raw_commands:
            cmd_type = cmd['command']

            if cmd_type == self.CommandType.AUTH:
                logger.debug('got auth cmd: %s', cmd)
                commands.append(AuthCommand(cmd, settings.get_config()['AUTH_FUNC']))

            elif cmd_type == self.CommandType.ACTION:
                logger.debug('got action: %s', cmd)
                action_type = cmd['action']['type']

                # subscribe actions
                if action_type == LOGUX_SUBSCRIBE:
                    channel = cmd['action']["channel"]
                    logger.debug('got subscription for channel: %s', channel)
                    commands.append(logux.channels[channel](cmd))
                    continue

                # custom actions
                if not logux.actions.has_action(action_type):
                    logger.warning('unknown action: %s', action_type)
                    commands.append(UnknownAction(cmd))
                    continue

                commands.append(logux.actions[action_type](cmd))

            else:
                logger.warning('wrong command type: %s', cmd)
                err_msg = f'wrong command type: {cmd_type}, expected {self.CommandType.choices}'
                logger.warning(err_msg)
                logger.warning('command with wrong type will be ignored')

        return commands

    def _is_server_authenticated(self) -> bool:
        """ Check Logux proxy server secret """
        return self.secret == settings.get_config()['CONTROL_SECRET']

    def apply_commands(self) -> Tuple[int, Union[str, LoguxValue]]:
        """ Apply all actions commands one by one

        :return: HTTP code and List of command applying results or error message
        """
        if not self._is_server_authenticated():
            # TODO: extract to common way to error response
            err_msg = 'Wrong secret'
            logger.warning(err_msg)
            return 403, err_msg

        if len(self.commands) == 0:
            return 200, [
                {
                    'answer': Command.ANSWER.ERROR,
                    'details': 'command list is empty'
                }
            ]

        return 200, list(filter(None, chain.from_iterable([cmd.apply() for cmd in self.commands])))


@csrf_exempt
@require_http_methods(["POST"])
def dispatch(request: HttpRequest):
    """ Entry point for all requests from Logux Proxy

    :param request: HTTP request from Logux Proxy server.

    :return: JSON response with results of commands applying
    """
    try:
        status, commands_results = LoguxRequest(request).apply_commands()
    except LoguxProxyException as err:
        status, commands_results = (400, str(err))

    if status != 200:
        r = HttpResponse(commands_results)
        r.status_code = status
        return r

    for cmd_res in commands_results:
        logger.debug(cmd_res)

    r = JsonResponse(commands_results, safe=False)
    r.status_code = status
    return r



class LoguxProxyException(Exception):
    """ Communication errors during acting with Logux Proxy server. """
    pass


class LoguxBadAuthException(Exception):
    """ Wrong AUTH command format or missing keys. """
    pass


class LoguxWrongLoadResultsException(Exception):
    """ load method of ChannelCommand returns invalid data. """
    pass



"""
    Django Logux integration engine: https://logux.io
"""
__title__ = 'Django Logux integration engine'
__version__ = "2.0.0rc3"
__author__ = 'Vadim Iskuchekov @egregors'
__license__ = 'MIT License'

# Synonyms
VERSION = __version__
AUTHOR = __author__

# Logux protocol version: https://logux.io/protocols/ws/spec/
LOGUX_PROTOCOL_VERSION = 4

default_app_config = 'logux.apps.LoguxConfig'
